"""
classes program which initializes and runs both the MQTT and InfluxDB controllers
"""

import logging
import signal
import threading
import time

from src.classes.common_classes import QueuePackage, SecretStore
from src.classes.influx_classes import InfluxConnector
from src.classes.mqtt_classes import MqttConnector
from src.helpers.consts import SOLAR_DEBUG_CONFIG_TITLE, THREADED_QUEUE
from src.helpers.py_logger import create_logger


class ThreadedRunner:
    """
    Class containing the threading operations to run the combined MQTT & Influx services
    """

    def __init__(self) -> None:
        self.log = create_logger(SOLAR_DEBUG_CONFIG_TITLE)
        self.thread_events = threading.Event()
        logging.logThreads = True

    def sigterm_handler(self, _signo, _stack_frame) -> None:
        """
        Handling SIGTERM signals
        """
        logging.critical("Received SIGTERM, shutting down")
        time.sleep(1)
        self.thread_events.clear()

    def sigint_handler(self, _signo, _stack_frame) -> None:
        """
        Handling SIGINT or CTRL + C signals
        """
        logging.critical("Received SIGINT/CTRL+C quit code, shutting down")
        time.sleep(1)
        self.thread_events.clear()

    def start(self) -> None:
        """
        Calls both the Influx database connector and the MQTT connector
        and runs them in separate threads
        NOTE: This program actually uses four threads not three due to the behavior
            of the MQTT loop_start() function
        """
        self.thread_events.set()
        logging.info("Created thread list")
        thread_list = [
            threading.Thread(
                name="Thread-Influx",
                target=self.run_threaded_influx_writer,
            ),
            threading.Thread(
                name="Thread-MQTT", target=self.run_threaded_mqtt_client, daemon=True
            ),
        ]

        signal.signal(signal.SIGTERM, self.sigterm_handler)
        signal.signal(signal.SIGINT, self.sigint_handler)

        # Starting threads
        logging.info("Starting threads")
        for thread in thread_list:
            thread.start()
            logging.info(f"Started thread: {thread.name}")

        # Put main thread to sleep
        logging.info("Main thread entering blocking loop")
        while self.thread_events.is_set():
            time.sleep(1)
        logging.info("Main thread exited blocking loop")

        # Gracefull terminate all threads
        logging.info("Clearing thread events, gracefully terminating all threads")
        self.thread_events.clear()

        # Closing threads
        for thread in thread_list:
            thread.join()
            logging.info(f"Joined thread: {thread.name}")
        logging.info("All threads have closed")
        logging.info("Exited application with exit code 0")

    def run_threaded_influx_writer(self) -> None:
        """
        Secondary thread which runs the InfluxDB connector
        Writes point data received from the MQTT._on_message in a threaded process
        NOTE: Since this program needs to indefinitely run all
        exceptions will just be logged instead of exiting the program.
        """
        try:
            secret_store = SecretStore(has_influx_access=True)
            influx_connector = InfluxConnector(secret_store=secret_store)
            logging.info("Attempting health check for InfluxDB")
        except Exception:
            logging.exception("Failed to setup environment")
            self.thread_events.clear()
            return
        try:
            influx_connector.health_check()
            logging.info("InfluxDB health check succeeded")
        except Exception:
            logging.exception("Failed to complete InfluxDB health check")
            self.thread_events.clear()
            return

        while self.thread_events.is_set():
            if not THREADED_QUEUE.empty():
                queue_package: QueuePackage = THREADED_QUEUE.get(timeout=1.0)
                logging.debug(
                    f"Popped packet off queue, queue now has {THREADED_QUEUE.qsize()} items"
                )
                try:
                    influx_connector.write_points(queue_package=queue_package)
                except Exception:
                    logging.exception(
                        "Failed to run write to Influx server, returned error"
                    )
                    time.sleep(1)
                if THREADED_QUEUE.empty():
                    logging.info("Popped all packets off queue and wrote to InfluxDB")
            else:
                time.sleep(0.5)
        self.thread_events.clear()

    def run_threaded_mqtt_client(self):
        """
        Main process which runs the MQTT connector
        Listens to a MQTT broken then decodes received packets
        NOTE: Since this program needs to indefinitely run all
        exceptions will just be logged instead of exiting the program.
        """
        secret_store = SecretStore(has_mqtt_access=True)
        mqtt_connector = MqttConnector(
            secret_store=secret_store,
        )
        mqtt_client = None
        logging.info("Creating MQTT listening service")
        try:
            mqtt_client = mqtt_connector.get_mqtt_client()
        except Exception:
            logging.exception("Failed to create MQTT listening service")
            self.thread_events.clear()
            return

        # Start MQTT-Listener thread to indefinitely listen to the broker
        # NOTE: This actually runs another thread as a Daemon due to the behavior of loop_start
        logging.info("Started thread: MQTT-Listener")
        # Running loop_start() will enter into a confection state and
        # enter on_message() every time a new message comes in. Because
        # of this all error/exception handing must be done in the on_message()
        # command, in our case we don't want the program to exit so we just log the error.
        try:
            mqtt_client.loop_start()
        except Exception:
            logging.exception("MQTT listener exited with a fatal error")
            self.thread_events.clear()
            return

        # Sleep Thread-MQTT
        while self.thread_events.is_set():
            time.sleep(1)

        # Stops the extra MQTT-Listener thread gracefully
        if mqtt_client:
            mqtt_client.loop_stop()
        logging.info("Joined thread: MQTT-Listener")
        self.thread_events.clear()


def main():
    """
    Main runtime for solar logger, called from start_logger.py
    """
    thread_runner = ThreadedRunner()
    thread_runner.start()



"""
File which contains definitions of custom exception patterns
"""


class MissingConfigurationError(Exception):
    """
    Defines an exception class for use when the loaded configuration file contains missing data
    """


class MissingCredentialsError(Exception):
    """
    Defines an exception class for when there are credentials missing
    """


class MqttServerOfflineError(Exception):
    """
    Defines an exception class for an offline MQTT server
    """



"""
Contains all functions required to setup logging
"""

import logging
import os
from configparser import ConfigParser
from logging import Logger, StreamHandler
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler

from src.classes.custom_exceptions import MissingConfigurationError
from src.helpers.consts import CONFIG_FILENAME, SOLAR_DEBUG_CONFIG_TITLE


class LoggingTools:
    """
    Class contains all tools required to create loggers
    """

    def __init__(
        self, config_name: str, logger: Logger, config_dir: str = CONFIG_FILENAME
    ) -> None:
        """
        Initialization of logging class
        """
        self._debug_level = None
        self._file_format = None
        self._date_format = None
        self._is_file_logging = None
        self._log_rotation = None
        self._file_location = None
        self._file_path = None
        self._max_file_bytes = None
        self._max_file_no = None
        self._mode = None
        self._logger = logger
        self._config_dir = config_dir
        self._config_name = config_name
        self._config_parser = ConfigParser()

    def create_loggers(self):
        """
        Creates loggers from the given configs
        """
        self._create_stdout_logger(logger=self._logger)
        if self._is_file_logging and self._log_rotation == "size_based":
            self._create_rotating_file_logger(logger=self._logger)
        elif self._is_file_logging and self._log_rotation == "time_based":
            self._create_timed_rotating_file_logger(logger=self._logger)

    def _read_basic_config(self) -> None:
        try:
            debug_dict = {
                "DEBUG": logging.DEBUG,
                "INFO": logging.INFO,
                "WARNING": logging.WARNING,
                "ERROR": logging.ERROR,
                "CRITICAL": logging.CRITICAL,
            }
            self._debug_level = debug_dict[
                self._config_parser.get(self._config_name, "debug_level")
            ]
            self._file_format = self._config_parser.get(self._config_name, "format")
            self._date_format = self._config_parser.get(self._config_name, "dateformat")
            self._is_file_logging = self._config_parser.getboolean(
                self._config_name, "file_logging"
            )
        except Exception as err:
            logging.critical("Failed to read basic logger configs")
            raise MissingConfigurationError(
                "Failed to read basic logger configs"
            ) from err

        if None in [self._debug_level, self._file_format, self._date_format]:
            logging.critical("Failed to read basic logger configs")
            raise MissingConfigurationError("Failed to read basic logger configs")

    def _read_extra_configs(self) -> None:
        try:
            self._log_rotation = self._config_parser.get(
                self._config_name, "log_rotation"
            )

            self._file_location = self._config_parser.get(
                self._config_name, "file_location"
            )
            file_name = self._config_parser.get(self._config_name, "file_name")
            self._file_path = self._file_location + file_name
            self._max_file_bytes = int(
                self._config_parser.get(self._config_name, "max_file_bytes")
            )
            self._max_file_no = int(
                self._config_parser.get(self._config_name, "max_file_no")
            )
            if self._config_name == SOLAR_DEBUG_CONFIG_TITLE:
                self._mode = str(self._config_parser.get(self._config_name, "mode"))
        except Exception as err:
            logging.critical("Failed to read file logger settings in configs")
            raise MissingConfigurationError(
                "Failed to read file logger settings in configs"
            ) from err

        if (
            None
            in [
                self._log_rotation,
                self._file_location,
                self._file_path,
                self._max_file_bytes,
                self._max_file_no,
            ]
            or (self._config_name == SOLAR_DEBUG_CONFIG_TITLE and self._mode is None)
            or self._mode in ["None", ""]
        ):
            logging.critical("Failed to read file logger settings in configs")
            raise MissingConfigurationError(
                "Failed to read file logger settings in configs"
            )

    def read_configs(self) -> None:
        """
        Reads config file and parses and stores the result
        """
        self._config_parser.read(self._config_dir)
        self._read_basic_config()
        if self._is_file_logging:
            self._read_extra_configs()

    def _create_stdout_logger(self, logger: Logger) -> None:
        """
        Creates a standard STDOUT logger
        """
        logger.setLevel(self._debug_level)
        stream_handler = StreamHandler()
        stream_handler.setLevel(self._debug_level)
        log_formatter = logging.Formatter(
            fmt=self._file_format, datefmt=self._date_format
        )
        stream_handler.setFormatter(log_formatter)
        logger.addHandler(stream_handler)
        logging.info("Created stdout logger")

    def _create_rotating_file_logger(self, logger: Logger) -> None:
        """
        Creates a rotating file logger which limits log files size
        and when exceeding that size, creates a new log file
        """
        if not os.path.exists(self._file_location):
            os.makedirs(self._file_location)  # pragma: no cover

        rotating_handler = RotatingFileHandler(
            filename=self._file_path,
            maxBytes=self._max_file_bytes,
            backupCount=self._max_file_no,
            mode=self._mode,
        )
        rotating_handler.setLevel(self._debug_level)
        log_formatter = logging.Formatter(
            fmt=self._file_format, datefmt=self._date_format
        )
        rotating_handler.setFormatter(log_formatter)
        logger.addHandler(rotating_handler)
        logging.info(f"Created rotating file log file at {self._file_path}")

    def _create_timed_rotating_file_logger(self, logger: Logger) -> None:
        """
        Creates a rotating file logger which limits log files size
        and when exceeding that size, creates a new log file
        """
        if not os.path.exists(self._file_location):
            os.makedirs(self._file_location)  # pragma: no cover
        rotating_time_handler = TimedRotatingFileHandler(
            filename=self._file_path,
            when="midnight",
            backupCount=self._max_file_no,
        )
        rotating_time_handler.suffix = "%Y-%m-%d"
        rotating_time_handler.setLevel(self._debug_level)
        log_formatter = logging.Formatter(
            fmt=self._file_format,
            datefmt=self._date_format,
        )
        rotating_time_handler.setFormatter(log_formatter)
        logger.addHandler(rotating_time_handler)
        logging.info(f"Created time rotating file log file at {self._file_path}")


def create_logger(
    config_name: str, config_dir: str = CONFIG_FILENAME
) -> Logger:  # pragma: no cover
    """
    Creates a logging instance, can be customized through the config.ini
    :param config_name: Section under the config for the configuration to pull data from
    :return: Logger for logging
    """

    logger = logging.getLogger()
    logging_tools = LoggingTools(
        config_name=config_name, config_dir=config_dir, logger=logger
    )
    logging_tools.read_configs()
    logging_tools.create_loggers()
    return logger



"""
Classes file, contains methods for the MQTT subscriber which will receive packets
from the broker and input them into an Influx database.
Check the Influx query documentation for write syntax:
https://docs.influxdata.com/influxdb/v2.0/api-guide/client-libraries/python/#query-data-from-influxdb-with-python
"""

import logging
import ssl
import struct
import time
from dataclasses import dataclass
from datetime import datetime
from typing import Tuple

from paho.mqtt.client import Client, MQTTMessage
from pymate.matenet import DCStatusPacket, FXStatusPacket, MXStatusPacket

from src.classes.common_classes import QueuePackage, SecretStore
from src.helpers.consts import QUEUE_WAIT_TIME, THREADED_QUEUE, TIME_PACKET_SIZE


class PyMateDecoder:
    """
    Class which decodes bytestreams received from the MQTT broker
    """

    @staticmethod
    def detach_time(
        msg: bytearray, padding_at_end: int = 0
    ) -> Tuple[bytearray, bytearray]:
        """
        Splits the bytearray into two objects, time and payload
        """
        raw_time = msg[:TIME_PACKET_SIZE]
        msg_time = struct.unpack("i", raw_time)
        msg_payload = msg[TIME_PACKET_SIZE:-padding_at_end]
        return msg_time[0], msg_payload

    @staticmethod
    def dc_decoder(msg: bytearray) -> dict:
        """
        Decoder for DC objects
        :param msg: Input message to decode
        :return: List of decoded objects
        """
        dc_packet = DCStatusPacket.from_buffer(msg).__dict__
        return {key: value for (key, value) in dc_packet.items() if key != "raw"}

    @staticmethod
    def fx_decoder(msg: bytearray) -> dict:
        """
        Decoder for FX objects
        :param msg: Input message to decode
        :return: List of decoded objects
        """
        fx_packet = FXStatusPacket.from_buffer(msg).__dict__
        return {key: value for (key, value) in fx_packet.items() if key != "raw"}

    @staticmethod
    def mx_decoder(msg: bytearray) -> dict:
        """
        Decoder for MX objects
        :param msg: Input message to decode
        :return: List of decoded objects
        """
        mx_packet = MXStatusPacket.from_buffer(msg).__dict__
        return {key: value for (key, value) in mx_packet.items() if key != "raw"}


@dataclass
class MqttTopics:
    """
    Object which is a model of all the different MQTT topics
    """

    mate_status = "mate/status"

    dc_name = "dc-1"
    dc_status = "mate/dc-1/status"
    dc_data = "mate/dc-1/dc-status"
    dc_raw = "mate/dc-1/stat/raw"
    dc_ts = "mate/dc-1/stat/ts"

    fx_name = "fx-1"
    fx_status = "mate/fx-1/status"
    fx_data = "mate/fx-1/fx-status"
    fx_raw = "mate/fx-1/stat/raw"
    fx_ts = "mate/fx-1/stat/ts"

    mx_name = "mx-1"
    mx_status = "mate/mx-1/status"
    mx_data = "mate/mx-1/mx-status"
    mx_raw = "mate/mx-1/stat/raw"
    mx_ts = "mate/mx-1/stat/ts"


class MqttConnector:
    """
    Class which creates a client to connect to MQTT subscriber and decode the messages
    """

    def __init__(
        self,
        secret_store: SecretStore,
    ) -> None:
        """
        :param host: Web url for the subscriber to listen on
        :param port: Port which the web server uses for MQTT
        :param user: Username to access MQTT server
        :param token: Token to access MQTT server
        :param influx_connector: Database for the MQTTDecoder to write results to
        """
        self._status = {
            MqttTopics.mate_status: "offline",
            MqttTopics.dc_status: "offline",
            MqttTopics.fx_status: "offline",
            MqttTopics.mx_status: "offline",
        }
        self._dec_msg = None
        self._mqtt_secrets = secret_store.mqtt_secrets
        self._mqtt_client = Client()

    @staticmethod
    def _on_socket_open(_client, userdata, sock) -> None:
        """
        Debug function for logging socket openings from MQTT
        """
        logging.debug(f"Socket open debug args, {userdata}, {sock}")

    @staticmethod
    def _on_socket_close(_client, userdata, sock) -> None:
        """
        Debug function for logging socket closes from MQTT
        """
        logging.debug(f"Socket close debug args, {userdata}, {sock}")

    @staticmethod
    def _on_subscribe(_client, userdata, mid, granted_qos) -> None:
        """
        Logs when the MQTT client calls on_subscribe
        """
        logging.info("Subscribed to MQTT topic, _on_subscribe")
        logging.info(f"MQTT topic returns QoS level of {granted_qos}")
        logging.debug(f"Subscribe debug args, {userdata}, {mid}, {granted_qos}")

    @staticmethod
    def _on_unsubscribe(_client, userdata, mid) -> None:
        """
        Logs when MQTT calls on_unsubscribe
        """
        logging.info("Unsubscribed from MQTT topic, _on_unsubscribe")
        logging.debug(f"Unsubscribe debug args, {userdata}, {mid}")

    def _on_connect(self, _client, userdata, flags, return_code) -> None:
        """
        Logs when MQTT calls on_connect
        The value of rc indicates success or not
        """
        return_codes = {
            0: "Connection successful",
            1: "Connection refused - incorrect protocol version",
            2: "Connection refused - invalid client identifier",
            3: "Connection refused - server unavailable",
            4: "Connection refused - bad username or password",
            5: "Connection refused - not authorized",
        }
        if return_code == 0:
            logging.info("Connecting to MQTT broker, _on_connect")
            topic = f"{self._mqtt_secrets['mqtt_topic']}"
            self._mqtt_client.subscribe(topic=topic)
        else:
            logging.error(
                f"Couldn't connect to MQTT broker returned code: {return_code}\n"
                f"{return_codes[return_code]}"
            )
            logging.debug(f"Connect debug args, {userdata}, {flags}, {return_code}")

    @staticmethod
    def _on_disconnect(_client, userdata, return_code) -> None:
        """
        Logs when MQTT calls on_disconnect
        """
        logging.warning("Disconnected from MQTT broker, _on_disconnect")
        logging.debug(f"Disconnect debug args, {userdata}, {return_code}")

    def _check_status(self, msg: MQTTMessage) -> None:
        """
        Called everytime a status message is received and checks the status of the server
        :param msg: Received message from MQTT broker
        """
        for topic, _ in self._status.items():
            if msg.topic == topic and msg.payload.decode("ascii") == "offline":
                self._status[topic] = "offline"
                logging.warning(f"{msg.topic} has gone offline")
            elif msg.topic == topic and msg.payload.decode("ascii") == "online":
                self._status[topic] = "online"
                logging.info(f"{msg.topic} is now online")

    @staticmethod
    def _load_queue(measurement: str, time_field: datetime, payload: dict) -> None:
        """
        Unpacks the payload and assigns a new message with each item in the payload,
        this sets all messages in the packet with the same time and measurement field
        Then loads each packet into a globally accessible queue.
        """
        for key, value in payload.items():
            # We don't like a queue building up since it means our program isn't
            # handling the volume of data or a service is offline
            while THREADED_QUEUE.full():
                logging.error(f"Queue is full, sleeping for {QUEUE_WAIT_TIME} seconds")
                time.sleep(QUEUE_WAIT_TIME)
            THREADED_QUEUE.put(
                QueuePackage(
                    measurement=measurement,
                    time_field=time_field,
                    field={key: float(value)},
                )
            )
        logging.info(
            f"Pushed items onto queue, queue now has {THREADED_QUEUE.qsize()} items"
        )

    def _decode_message(self, msg: MQTTMessage) -> None:
        """
        Handles all code around decoding raw bytestrings and loading the packets into a global queue
        :param msg: Takes in a raw bytestring from MQTT
        """
        dc_online = self._status[MqttTopics.dc_status]
        fx_online = self._status[MqttTopics.fx_status]
        mx_online = self._status[MqttTopics.mx_status]

        if msg.topic == MqttTopics.dc_data and dc_online == "online":
            logging.info(f"Received {MqttTopics.dc_name} data packet")
            logging.debug(f"{MqttTopics.dc_name} payload: {msg.payload}")
            # NOTE: Due to errors in our packet packing, it introduces a random buffer at the end
            padding_at_end = 2
            msg_time, msg_payload = PyMateDecoder.detach_time(
                msg=msg.payload, padding_at_end=padding_at_end
            )
            dc_time = datetime.fromtimestamp(msg_time)
            dc_payload = PyMateDecoder.dc_decoder(msg_payload)
            logging.debug(
                f"Decoded and split {MqttTopics.dc_name} payload: {dc_payload} at {dc_time}"
            )
            self._load_queue(
                measurement=MqttTopics.dc_name, time_field=dc_time, payload=dc_payload
            )

        if msg.topic == MqttTopics.fx_data and fx_online == "online":
            logging.info(f"Received {MqttTopics.fx_name} data packet")
            logging.debug(f"{MqttTopics.fx_name} payload: {msg.payload}")
            # NOTE: Due to errors in our packet packing, it introduces a random buffer at the end
            padding_at_end = 3
            msg_time, msg_payload = PyMateDecoder.detach_time(
                msg=msg.payload, padding_at_end=padding_at_end
            )
            fx_time = datetime.fromtimestamp(msg_time)
            fx_payload = PyMateDecoder.fx_decoder(msg_payload)
            logging.debug(
                f"Decoded and split {MqttTopics.fx_name} payload: {fx_payload} at {fx_time}"
            )
            self._load_queue(
                measurement=MqttTopics.fx_name, time_field=fx_time, payload=fx_payload
            )

        if msg.topic == MqttTopics.mx_data and mx_online == "online":
            logging.info(f"Received {MqttTopics.mx_name} data packet")
            logging.debug(f"{MqttTopics.mx_name} payload: {msg.payload}")
            # NOTE: Due to errors in our packet packing, it introduces a random buffer at the end
            padding_at_end = 3
            msg_time, msg_payload = PyMateDecoder.detach_time(
                msg=msg.payload, padding_at_end=padding_at_end
            )
            mx_time = datetime.fromtimestamp(msg_time)
            mx_payload = PyMateDecoder.mx_decoder(msg_payload)
            logging.debug(
                f"Decoded and split {MqttTopics.mx_name} payload: {mx_payload} at {mx_time}"
            )
            self._load_queue(
                measurement=MqttTopics.mx_name, time_field=mx_time, payload=mx_payload
            )

    def _on_message(self, _client, _userdata, msg: MQTTMessage) -> None:
        """
        Called everytime a message is received which it then decodes
        :param msg: Message to partition into categories and decode
        """
        try:
            self._check_status(msg=msg)
            if self._status[MqttTopics.mate_status] == "online":
                self._decode_message(msg=msg)
            else:
                logging.warning(f"{MqttTopics.mate_status} is offline")
        except Exception:
            logging.exception("MQTT on_message raised an exception:")

    def get_mqtt_client(self) -> Client:
        """
        Initial setup for the MQTT connector, connects to MQTT broker
        and failing to connect will exit the program
        """
        self._mqtt_client.username_pw_set(
            username=self._mqtt_secrets["mqtt_user"],
            password=self._mqtt_secrets["mqtt_token"],
        )

        self._mqtt_client.on_connect = self._on_connect
        self._mqtt_client.on_message = self._on_message
        self._mqtt_client.on_disconnect = self._on_disconnect
        self._mqtt_client.on_unsubscribe = self._on_unsubscribe
        self._mqtt_client.on_subscribe = self._on_subscribe
        self._mqtt_client.on_socket_open = self._on_socket_open
        self._mqtt_client.on_socket_close = self._on_socket_close

        self._mqtt_client.tls_set(cert_reqs=ssl.CERT_NONE)
        self._mqtt_client.tls_insecure_set(True)

        _ = self._mqtt_client.connect(
            host=self._mqtt_secrets["mqtt_host"],
            port=self._mqtt_secrets["mqtt_port"],
        )

        return self._mqtt_client



"""
Classes file, contains methods for the Influx database controller
to do writes and queries to the database
"""

import logging
from datetime import datetime

from influxdb_client import InfluxDBClient
from influxdb_client.client.write_api import SYNCHRONOUS

from src.classes.common_classes import QueuePackage, SecretStore


class InfluxConnector:
    """
    Class which creates a client to access and modify a connected database
    """

    def __init__(self, secret_store: SecretStore) -> None:
        """
        :param token: Secret password to login to database with
        :param org: Organization of the bucket to login to
        :param bucket: Database source
        :param url: Web address to connect to database
        """
        _influx_secrets = secret_store.influx_secrets

        self._influx_org = _influx_secrets["influx_org"]
        self._influx_bucket = _influx_secrets["influx_bucket"]

        logging.info("Initializing InfluxDB client")
        self._influx_client = InfluxDBClient(
            url=_influx_secrets["influx_url"],
            org=_influx_secrets["influx_org"],
            token=_influx_secrets["influx_token"],
        )
        logging.info("Initializing Influx write api")
        self._write_client = self._influx_client.write_api(write_options=SYNCHRONOUS)
        logging.info("Initializing Influx query api")
        self._query_client = self._influx_client.query_api(query_options=SYNCHRONOUS)

    def health_check(self) -> None:
        """
        Defines the initialization of the Influx connector,
        invoking the connection to the InfluxDB and write API
        """
        self._influx_client.ready()  # External request

    @staticmethod
    def _verify_queue_package(queue_package: QueuePackage):
        assertion_message = "The received queue_packed has malformed data: "
        assert queue_package is not None, assertion_message + "queue_package empty"
        assert isinstance(queue_package.measurement, str), (
            assertion_message + "type of measurement not str"
        )
        assert isinstance(queue_package.field, dict | str), (
            assertion_message + "type of field not, dict | str"
        )
        assert isinstance(queue_package.time_field, datetime), (
            assertion_message + "type of time_field not, datetime"
        )

    def write_points(self, queue_package: QueuePackage) -> None:
        """
        Writes points to InfluxDB
        :param msg_time: Time value of the packet
        :param msg_type: Type of header the msg carries, either FX, MX or DX
        :param msg_payload: Dictionary of messages for MQTTDecoder to input into the
            Influx Database in dictionary
        """
        self._verify_queue_package(queue_package=queue_package)
        self._write_client.write(
            bucket=self._influx_bucket,
            org=self._influx_org,
            record={
                "measurement": queue_package.measurement,
                "fields": queue_package.field,
            },
            time=queue_package.time_field,
        )  # External request
        logging.debug(f"Wrote point: {queue_package} at {queue_package.time_field}")

    def query_database(self, query_mode: str, query: str) -> None:
        """
        Runs given query on Influx database and returns results
        :param query_mode: Defines what mode to run the query in,
            supports "csv", "flux" and "stream"
        :param query: Input query to run on Influx database
        """
        query_result = None
        if query_mode == "csv":
            query_result = self._query_client.query_csv(
                org=self._influx_org, query=query
            )  # External request
        elif query_mode == "flux":
            query_result = self._query_client.query(
                org=self._influx_org, query=query
            )  # External request
        elif query_mode == "stream":
            query_result = self._query_client.query_stream(
                org=self._influx_org, query=query
            )  # External request
        logging.debug("Query to Influx server was successful")
        return query_result



"""
Contains all functions that aren't directly correlated to Influx, MQTT, or logging
"""

import csv
import logging
import os
from configparser import ConfigParser

from src.helpers.consts import CONFIG_FILENAME


def write_results_to_csv(config_name: str, table: dict) -> None:
    """
    Writes a CSV file from an Influx query
    :param config_name: Section under the config for the configuration to pull data from
    :param table: Resultant CSV query from the Influx database
    """
    try:
        config_parser = ConfigParser()
        config_parser.read(CONFIG_FILENAME)
        file_location = config_parser.get(config_name, "csv_location")
        filename = config_parser.get(config_name, "csv_name")
        full_path = file_location + filename
        filemode = config_parser.get(config_name, "csv_mode")
        if not os.path.exists(file_location):
            os.makedirs(file_location)
        with open(full_path, filemode) as file_instance:
            writer = csv.writer(file_instance)
            for row in table:
                writer.writerow(row)
        logging.info(f"Wrote rows into CSV file at: {full_path}")
    except Exception as err:
        logging.critical("Failed to write CSV")
        raise err


def read_query_settings(config_name: str) -> any:
    """
    :param config_name: Section under the config for the configuration to pull data from
    :return: Query variables
    """
    config_parser = ConfigParser()
    config_parser.read(CONFIG_FILENAME)
    return config_parser.get(section=config_name, option="query_mode")



# pylint: disable=missing-function-docstring, missing-module-docstring, missing-class-docstring
"""
Setup instructions for package management
"""
import re

import setuptools

PACKAGE_NAME = "solar_logger"
PACKAGE_DIR = "src"
EXCLUDED_PACKAGES = ["*tests*"]

with open("README.md", "r", encoding="utf-8") as open_readme_file:
    long_description = open_readme_file.read()


def get_version():
    with open("src/__version__.py", "r") as open_version_file:
        version = open_version_file.read()
        version = version.strip().split(" = ")[1]
        version = re.sub("[\"']", "", version)
        return re.sub("[-*]", "_", version)


VERSION = get_version()

setuptools.setup(
    name=PACKAGE_NAME.lower(),
    version=VERSION,
    author="Zach Sanson",
    author_email="zac@sanson.co.nz",
    description="MQTT to Influx DB converter for Outback solar controller",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/WibblyGhost/solar_logger",
    project_urls={
        "Bug Tracker": "https://github.com/WibblyGhost/solar_logger/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_dir={"": PACKAGE_DIR},
    packages=setuptools.find_packages(where=PACKAGE_DIR, exclude=EXCLUDED_PACKAGES),
    python_requires=">=3.10",
    install_requires=[
        "influxdb_client==1.20.0",
        "paho_mqtt==1.5.1",
        "pymate @ git+https://github.com/WibblyGhost/pymate.git@feature/python3",
        "Rx~=3.2.0",
        "pytz~=2021.1",
    ],
)



"""
Supporting classes for querying a InfluxDB server is contained here
"""
import logging


class QueryBuilder:
    """
    Class which creates a query to send to the Influx database
    """

    query_string = None

    def __init__(self, bucket: str, start_range: str, end_range: str = None) -> None:
        """
        Creates a base string for the query from which can be built upon
        :param bucket:  Influx database bucket for query
        :param start_range: The earliest time to include in results
        :param end_range: The latest time to include in results, defaults to now()
        """
        self._filter_field = ""
        self._aggregate_field = ""
        self._sort_field = ""
        self._bucket = bucket
        self._start_range = start_range
        self._end_range = end_range

    def __str__(self) -> str:
        """
        :return: String representation of the query
        """
        return self._build_string()

    def __repr__(self) -> str:
        """
        :return: Raw representation of the query
        """
        return repr(self.__str__())

    @staticmethod
    def help() -> None:
        """
        Prints how to use QueryBuilder for the python interactive users.
        """
        print(
            """
        QueryBuilder(bucket, start_range, end_range)
            Creates a base string for the query from which can be built upon
        :param bucket:  Influx database bucket for query
        :param start_range: The earliest time to include in results
        :param end_range=None: The latest time to include in results, defaults to now()

        QueryBuilder.append_filter(self, field_1, value_1, joiner, new_band)
            Adds filter fields to the query, function is repeatable and
            can therefore add multiple filters
        :param new_band: If true, creates a new filter field instead of appending the filter field
        :param field_1: Takes _measurement, _tag or _field
        :param value_1: Value you want the field to equal
        :param joiner: Optional join operator, can be "And" / "Or"

        QueryBuilder.append_aggregate(self, collection_window, aggregate_function)
                Adds an aggregation field to the query
        :param collection_window: Time frame for the data to aggregate
        :param aggregate_function: What function to apply to the window

        QueryBuilder.append_sort(self, field, desc)
                Adds a sort field to a query
        :param field: Field to sort results by
        :param desc=False: Ascending or descending
        """
        )

    def _build_string(self) -> str:
        """
        Creates basic string from set function variables
        :return: Built query in string form
        """
        self.query_string = self._append_from
        self.query_string += self._append_time_range
        self.query_string += self._filter_field
        self.query_string += self._aggregate_field
        self.query_string += self._sort_field
        logging.debug(f"Built query string:\n{self.query_string}")
        return self.query_string

    @property
    def _append_from(self) -> str:
        """
        Adds from field to query, takes bucket attribute and appends to classes string
        :param self.bucket: Influx database bucket to query
        """
        logging.debug("Created query from field")
        return f'from(bucket: "{self._bucket}")'  # Must use single quotes

    @property
    def _append_time_range(self) -> str:
        """
        Adds time range to query, takes start range and optional end range
        Can use queries like "-10m" or datetime stamps
        :param self.start_range: The earliest time to include in results
        :param self.end_range: The latest time to include in results, defaults to now()
        """
        logging.debug("Created query time range field")
        if self._end_range:
            return f"\n\t|> range(start: {self._start_range}, stop: {self._end_range})"
        return f"\n\t|> range(start: {self._start_range})"

    def append_filter(
        self, field: str, value: str, joiner: str = None, new_band: bool = False
    ) -> None:
        """
        Adds filter fields to the query, function is repeatable and
        can therefore add multiple filters
        :param new_band: If true, creates a new filter field instead of appending the filter field
        :param field_1: Takes _measurement, _tag or _field
        :param value_1: Value you want the field to equal
        :param joiner: Optional join operator, can be "And" / "Or"
        """
        logging.debug("Created query filter field")
        if not self._filter_field or new_band:
            self._filter_field += "\n\t|> filter(fn: (r) => "
        self._filter_field += f'r["{field}"] == "{value}")'  # Must use single quotes
        if joiner:
            self._filter_field = self._filter_field[:-1]
            self._filter_field += f" {joiner} "

    def append_aggregate(self, collection_window: str, aggregate_function: str) -> None:
        """
        Adds an aggregation field to the query
        :param collection_window: Time frame for the data to aggregate
        :param aggregate_function: What function to apply to the window
        """
        logging.debug("Created query aggregate field")
        self._aggregate_field = (
            f"\n\t|> aggregateWindow(every:"
            f" {collection_window}, fn: {aggregate_function}"
        )

    def append_sort(self, field: str, desc: bool = False) -> None:
        """
        Adds a sort field to a query
        :param field: Field to sort results by
        :param desc: Ascending or descending
        """
        logging.debug("Created query sort field")
        self._sort_field = (
            f'\n\t|> sort(columns: ["{field}"], desc: {desc}'  # Must use single quotes
        )
